@@ -27,6 +27,7 @@
 #include <config.h>
 #include <glib/gi18n-lib.h>
 #include <string.h>
+#include <math.h>
 #include <X11/Xlib.h>
 
 #include <X11/extensions/Xrandr.h>
@@ -48,6 +49,10 @@
 
 #define SERVERS_RANDR_IS_AT_LEAST_1_3(priv) (priv->rr_major_version > 1 || (priv->rr_major_version == 1 && priv->rr_minor_version >= 3))
 
+#define INTERFACE_SETTINGS "org.cinnamon.desktop.interface"
+#define GLOBAL_SCALE_FACTOR_KEY "scaling-factor"
+#define USE_UPSCALING_KEY "upscale-fractional-scaling"
+
 enum {
     SCREEN_PROP_0,
     SCREEN_PROP_GDK_SCREEN,
@@ -99,6 +104,7 @@ struct GnomeRRCrtc
     GnomeRROutput **	possible_outputs;
     int			x;
     int			y;
+    float       scale;
     
     GnomeRRRotation	current_rotation;
     GnomeRRRotation	rotations;
@@ -113,6 +119,9 @@ struct GnomeRRMode
     int			width;
     int			height;
     int			freq;		/* in mHz */
+    gboolean    doublescan;
+    gboolean    interlaced;
+    gboolean    vsync;
 };
 
 /* GnomeRRCrtc */
@@ -280,13 +289,13 @@ has_similar_mode (GnomeRROutput *output, GnomeRRMode *mode)
 
     for (i = 0; modes[i] != NULL; ++i)
     {
-	GnomeRRMode *m = modes[i];
+        GnomeRRMode *m = modes[i];
 
-	if (gnome_rr_mode_get_width (m) == width	&&
-	    gnome_rr_mode_get_height (m) == height)
-	{
-	    return TRUE;
-	}
+        if (gnome_rr_mode_get_width (m) == width	&&
+            gnome_rr_mode_get_height (m) == height)
+        {
+            return TRUE;
+        }
     }
 
     return FALSE;
@@ -300,42 +309,42 @@ gather_clone_modes (ScreenInfo *info)
 
     for (i = 0; info->outputs[i] != NULL; ++i)
     {
-	int j;
-	GnomeRROutput *output1, *output2;
+        int j;
+        GnomeRROutput *output1, *output2;
 
-	output1 = info->outputs[i];
-	
-	if (!output1->connected)
-	    continue;
-	
-	for (j = 0; output1->modes[j] != NULL; ++j)
-	{
-	    GnomeRRMode *mode = output1->modes[j];
-	    gboolean valid;
-	    int k;
+        output1 = info->outputs[i];
 
-	    valid = TRUE;
-	    for (k = 0; info->outputs[k] != NULL; ++k)
-	    {
-		output2 = info->outputs[k];
-		
-		if (!output2->connected)
-		    continue;
-		
-		if (!has_similar_mode (output2, mode))
-		{
-		    valid = FALSE;
-		    break;
-		}
-	    }
+        if (!output1->connected)
+            continue;
 
-	    if (valid)
-		g_ptr_array_add (result, mode);
-	}
+        for (j = 0; output1->modes[j] != NULL; ++j)
+        {
+            GnomeRRMode *mode = output1->modes[j];
+            gboolean valid;
+            int k;
+
+            valid = TRUE;
+            for (k = 0; info->outputs[k] != NULL; ++k)
+            {
+                output2 = info->outputs[k];
+
+                if (!output2->connected)
+                    continue;
+
+                if (!has_similar_mode (output2, mode))
+                {
+                    valid = FALSE;
+                    break;
+                }
+            }
+
+            if (valid)
+                g_ptr_array_add (result, mode);
+        }
     }
 
     g_ptr_array_add (result, NULL);
-    
+
     info->clone_modes = (GnomeRRMode **)g_ptr_array_free (result, FALSE);
 }
 
@@ -348,60 +357,63 @@ fill_screen_info_from_resources (ScreenInfo *info,
     GPtrArray *a;
     GnomeRRCrtc **crtc;
     GnomeRROutput **output;
-
     info->resources = resources;
 
     /* We create all the structures before initializing them, so
      * that they can refer to each other.
      */
     a = g_ptr_array_new ();
+
     for (i = 0; i < resources->ncrtc; ++i)
     {
-	GnomeRRCrtc *crtc = crtc_new (info, resources->crtcs[i]);
+        GnomeRRCrtc *crtc = crtc_new (info, resources->crtcs[i]);
 
-	g_ptr_array_add (a, crtc);
+        g_ptr_array_add (a, crtc);
     }
+
     g_ptr_array_add (a, NULL);
     info->crtcs = (GnomeRRCrtc **)g_ptr_array_free (a, FALSE);
 
     a = g_ptr_array_new ();
     for (i = 0; i < resources->noutput; ++i)
     {
-	GnomeRROutput *output = output_new (info, resources->outputs[i]);
+        GnomeRROutput *output = output_new (info, resources->outputs[i]);
 
-	g_ptr_array_add (a, output);
+        g_ptr_array_add (a, output);
     }
+
     g_ptr_array_add (a, NULL);
     info->outputs = (GnomeRROutput **)g_ptr_array_free (a, FALSE);
 
     a = g_ptr_array_new ();
     for (i = 0;  i < resources->nmode; ++i)
     {
-	GnomeRRMode *mode = mode_new (info, resources->modes[i].id);
+        GnomeRRMode *mode = mode_new (info, resources->modes[i].id);
 
-	g_ptr_array_add (a, mode);
+        g_ptr_array_add (a, mode);
     }
+
     g_ptr_array_add (a, NULL);
     info->modes = (GnomeRRMode **)g_ptr_array_free (a, FALSE);
 
     /* Initialize */
     for (crtc = info->crtcs; *crtc; ++crtc)
     {
-	if (!crtc_initialize (*crtc, resources, error))
-	    return FALSE;
+        if (!crtc_initialize (*crtc, resources, error))
+            return FALSE;
     }
 
     for (output = info->outputs; *output; ++output)
     {
-	if (!output_initialize (*output, resources, error))
-	    return FALSE;
+        if (!output_initialize (*output, resources, error))
+            return FALSE;
     }
 
     for (i = 0; i < resources->nmode; ++i)
     {
-	GnomeRRMode *mode = mode_by_id (info, resources->modes[i].id);
+        GnomeRRMode *mode = mode_by_id (info, resources->modes[i].id);
 
-	mode_initialize (mode, &(resources->modes[i]));
+        mode_initialize (mode, &(resources->modes[i]));
     }
 
     gather_clone_modes (info);
@@ -734,6 +746,8 @@ gnome_rr_screen_initable_init (GInitable *initable, GCancellable *canc, GError *
     int event_base;
     int ignore;
 
+    priv->interface_settings = g_settings_new ("org.cinnamon.desktop.interface");
+
     priv->connector_type_atom = XInternAtom (dpy, "ConnectorType", FALSE);
 
     if (XRRQueryExtension (dpy, &event_base, &ignore))
@@ -788,6 +802,8 @@ gnome_rr_screen_finalize (GObject *gobject)
     if (screen->priv->info)
       screen_info_free (screen->priv->info);
 
+    g_clear_object (&screen->priv->interface_settings);
+
     G_OBJECT_CLASS (gnome_rr_screen_parent_class)->finalize (gobject);
 }
 
@@ -990,6 +1006,8 @@ gnome_rr_screen_set_size (GnomeRRScreen *screen,
 {
     g_return_if_fail (GNOME_IS_RR_SCREEN (screen));
 
+    g_debug ("Setting screen size: %d x %d, %dmm x %dmm", width, height, mm_width, mm_height);
+
     gdk_error_trap_push ();
     XRRSetScreenSize (screen->priv->xdisplay, screen->priv->xroot,
 		      width, height, mm_width, mm_height);
@@ -1560,8 +1578,9 @@ out:
 static gboolean
 output_initialize (GnomeRROutput *output, XRRScreenResources *res, GError **error)
 {
-    XRROutputInfo *info = XRRGetOutputInfo (
-	DISPLAY (output), res, output->id);
+    XRROutputInfo *info = XRRGetOutputInfo (DISPLAY (output),
+                                            res,
+                                            output->id);
     GPtrArray *a;
     int i;
     
@@ -1571,14 +1590,14 @@ output_initialize (GnomeRROutput *output, XRRScreenResources *res, GError **erro
     
     if (!info || !output->info)
     {
-	/* FIXME: see the comment in crtc_initialize() */
-	/* Translators: here, an "output" is a video output */
-	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
-		     _("could not get information about output %d"),
-		     (int) output->id);
-	return FALSE;
+        /* FIXME: see the comment in crtc_initialize() */
+        /* Translators: here, an "output" is a video output */
+        g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
+                     _("could not get information about output %d"),
+                     (int) output->id);
+        return FALSE;
     }
-    
+
     output->name = g_strdup (info->name); /* FIXME: what is nameLen used for? */
     output->current_crtc = crtc_by_id (output->info, info->crtc);
     output->width_mm = info->mm_width;
@@ -1591,38 +1610,38 @@ output_initialize (GnomeRROutput *output, XRRScreenResources *res, GError **erro
     
     for (i = 0; i < info->ncrtc; ++i)
     {
-	GnomeRRCrtc *crtc = crtc_by_id (output->info, info->crtcs[i]);
-	
-	if (crtc)
-	    g_ptr_array_add (a, crtc);
+        GnomeRRCrtc *crtc = crtc_by_id (output->info, info->crtcs[i]);
+
+        if (crtc)
+            g_ptr_array_add (a, crtc);
     }
     g_ptr_array_add (a, NULL);
     output->possible_crtcs = (GnomeRRCrtc **)g_ptr_array_free (a, FALSE);
-    
+
     /* Clones */
     a = g_ptr_array_new ();
     for (i = 0; i < info->nclone; ++i)
     {
-	GnomeRROutput *gnome_rr_output = gnome_rr_output_by_id (output->info, info->clones[i]);
-	
-	if (gnome_rr_output)
-	    g_ptr_array_add (a, gnome_rr_output);
+        GnomeRROutput *gnome_rr_output = gnome_rr_output_by_id (output->info, info->clones[i]);
+
+        if (gnome_rr_output)
+            g_ptr_array_add (a, gnome_rr_output);
     }
     g_ptr_array_add (a, NULL);
     output->clones = (GnomeRROutput **)g_ptr_array_free (a, FALSE);
-    
+
     /* Modes */
     a = g_ptr_array_new ();
     for (i = 0; i < info->nmode; ++i)
     {
-	GnomeRRMode *mode = mode_by_id (output->info, info->modes[i]);
-	
-	if (mode)
-	    g_ptr_array_add (a, mode);
+        GnomeRRMode *mode = mode_by_id (output->info, info->modes[i]);
+
+        if (mode)
+            g_ptr_array_add (a, mode);
     }
     g_ptr_array_add (a, NULL);
     output->modes = (GnomeRRMode **)g_ptr_array_free (a, FALSE);
-    
+
     output->n_preferred = info->npreferred;
     
     /* Edid data */
@@ -2135,6 +2154,62 @@ xrotation_from_rotation (GnomeRRRotation r)
     return result;
 }
 
+static void
+set_crtc_scale (GnomeRRCrtc *crtc,
+                GnomeRRMode *mode,
+                float        scale,
+                guint        global_scale)
+{
+    gchar *filter;
+    float real_scale;
+    int i;
+    int looks_like_w, looks_like_h;
+
+    real_scale = global_scale / scale;
+
+    looks_like_w = 0;
+    looks_like_h = 0;
+
+    if (mode != NULL)
+    {
+        looks_like_w = gnome_rr_mode_get_width (mode) * real_scale / global_scale;
+        looks_like_h = gnome_rr_mode_get_height (mode) * real_scale / global_scale;
+    }
+
+    // Fractional scales use bilinear, doubling/halving can use nearest for better quality.
+    if ((XDoubleToFixed (real_scale) % 32768) > 0)
+    {
+        filter = g_strdup ("bilinear");
+    }
+    else
+    {
+        filter = g_strdup ("nearest");
+    }
+
+    g_debug ("\n\n(xid: %lu) Transforming based on:\n"
+             "global ui scale: %d\n"
+             "requested logical scale: %.2f\n"
+             "requested logical size: %dx%d\n"
+             "xrandr transform value: %.2f (%d)\n"
+             "scaling method: %s",
+             crtc->id,
+             global_scale, scale,
+             looks_like_w, looks_like_h,
+             real_scale, XDoubleToFixed (real_scale),
+             filter);
+
+    XTransform transform =  {{
+                    { XDoubleToFixed (real_scale), 0                          , 0                    },
+                    { 0                          , XDoubleToFixed (real_scale), 0                    },
+                    { 0                          , 0                          , XDoubleToFixed (1.0) },
+    }};
+
+    XRRSetCrtcTransform (DISPLAY (crtc), crtc->id,
+                         &transform,
+                         filter,
+                         NULL, 0);
+}
+
 gboolean
 gnome_rr_crtc_set_config_with_time (GnomeRRCrtc      *crtc,
 				    guint32           timestamp,
@@ -2144,6 +2219,8 @@ gnome_rr_crtc_set_config_with_time (GnomeRRCrtc      *crtc,
 				    GnomeRRRotation   rotation,
 				    GnomeRROutput   **outputs,
 				    int               n_outputs,
+                    float             scale,
+                    guint             global_scale,
 				    GError          **error)
 {
     ScreenInfo *info;
@@ -2160,21 +2237,20 @@ gnome_rr_crtc_set_config_with_time (GnomeRRCrtc      *crtc,
     
     if (mode)
     {
-	if (x + mode->width > info->max_width
-	    || y + mode->height > info->max_height)
-	{
-	    g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_BOUNDS_ERROR,
-			 /* Translators: the "position", "size", and "maximum"
-			  * words here are not keywords; please translate them
-			  * as usual.  A CRTC is a CRT Controller (this is X terminology) */
-			 _("requested position/size for CRTC %d is outside the allowed limit: "
-			   "position=(%d, %d), size=(%d, %d), maximum=(%d, %d)"),
-			 (int) crtc->id,
-			 x, y,
-			 mode->width, mode->height,
-			 info->max_width, info->max_height);
-	    return FALSE;
-	}
+        if (x + mode->width > info->max_width || y + mode->height > info->max_height)
+        {
+            g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_BOUNDS_ERROR,
+                         /* Translators: the "position", "size", and "maximum"
+                          * words here are not keywords; please translate them
+                          * as usual.  A CRTC is a CRT Controller (this is X terminology) */
+                         _("requested position/size for CRTC %d is outside the allowed limit: "
+                         "position=(%d, %d), size=(%d, %d), maximum=(%d, %d)"),
+                         (int) crtc->id,
+                         x, y,
+                         mode->width, mode->height,
+                         info->max_width, info->max_height);
+            return FALSE;
+        }
     }
     
     output_ids = g_array_new (FALSE, FALSE, sizeof (RROutput));
@@ -2185,7 +2261,9 @@ gnome_rr_crtc_set_config_with_time (GnomeRRCrtc      *crtc,
 	    g_array_append_val (output_ids, outputs[i]->id);
     }
 
-    gdk_error_trap_push ();   
+    gdk_error_trap_push ();
+
+    set_crtc_scale (crtc, mode, scale, global_scale);
     status = XRRSetCrtcConfig (DISPLAY (crtc), info->resources, crtc->id,
 			       timestamp, 
 			       x, y,
@@ -2260,6 +2338,14 @@ gnome_rr_crtc_get_position (GnomeRRCrtc *crtc,
 	*y = crtc->y;
 }
 
+float
+gnome_rr_crtc_get_scale (GnomeRRCrtc           *crtc)
+{
+    g_return_val_if_fail (crtc != NULL, MINIMUM_LOGICAL_SCALE_FACTOR);
+
+    return crtc->scale;
+}
+
 /* FIXME: merge with get_mode()? */
 GnomeRRRotation
 gnome_rr_crtc_get_current_rotation (GnomeRRCrtc *crtc)
@@ -2327,6 +2413,88 @@ crtc_copy (const GnomeRRCrtc *from)
     return to;
 }
 
+static float
+scale_from_transformation (XRRCrtcTransformAttributes *transformation)
+{
+  XTransform *xt;
+  float scale;
+
+  if (!transformation)
+    return 1.0f;
+
+  xt = &transformation->currentTransform;
+
+  if (xt->matrix[0][0] == xt->matrix[1][1])
+    scale = XFixedToDouble (xt->matrix[0][0]);
+  else
+    scale = XFixedToDouble (xt->matrix[0][0] + xt->matrix[1][1]) / 2.0;
+
+  return 1.0f / scale;
+}
+
+static float
+get_transform_scale (GnomeRRCrtc *crtc)
+{
+    XRRCrtcTransformAttributes *transform_attributes;
+    float ret;
+
+    if (!XRRGetCrtcTransform (DISPLAY (crtc), crtc->id, &transform_attributes))
+    {
+        transform_attributes = NULL;
+    }
+
+    ret = scale_from_transformation (transform_attributes);
+
+    XFree (transform_attributes);
+
+    return ret;
+}
+
+guint
+gnome_rr_screen_get_global_scale (GnomeRRScreen *screen)
+{
+    GdkScreen *gdkscreen;
+    GValue value = G_VALUE_INIT;
+    gint window_scale = 1;
+
+    gdkscreen = gdk_screen_get_default ();
+
+    g_value_init (&value, G_TYPE_INT);
+
+    if (gdk_screen_get_setting (gdkscreen, "gdk-window-scaling-factor", &value))
+    {
+        window_scale = g_value_get_int (&value);
+    }
+
+    return (guint) CLAMP (window_scale, MINIMUM_GLOBAL_SCALE_FACTOR, MAXIMUM_GLOBAL_SCALE_FACTOR);
+}
+
+guint
+gnome_rr_screen_get_global_scale_setting (GnomeRRScreen *screen)
+{
+    guint scale_factor;
+
+    scale_factor = g_settings_get_uint (screen->priv->interface_settings,
+                                        GLOBAL_SCALE_FACTOR_KEY);
+
+    return scale_factor;
+}
+
+void
+gnome_rr_screen_set_global_scale_setting (GnomeRRScreen *screen,
+                                          guint          scale_factor)
+{
+    g_settings_set_uint (screen->priv->interface_settings,
+                         GLOBAL_SCALE_FACTOR_KEY,
+                         scale_factor);
+}
+
+static float
+get_crtc_scale (GnomeRRCrtc *crtc)
+{
+    return gnome_rr_screen_get_global_scale (NULL) * get_transform_scale (crtc);
+}
+
 static gboolean
 crtc_initialize (GnomeRRCrtc        *crtc,
 		 XRRScreenResources *res,
@@ -2342,16 +2510,16 @@ crtc_initialize (GnomeRRCrtc        *crtc,
     
     if (!info)
     {
-	/* FIXME: We need to reaquire the screen resources */
-	/* FIXME: can we actually catch BadRRCrtc, and does it make sense to emit that? */
+        /* FIXME: We need to reaquire the screen resources */
+        /* FIXME: can we actually catch BadRRCrtc, and does it make sense to emit that? */
 
-	/* Translators: CRTC is a CRT Controller (this is X terminology).
-	 * It is *very* unlikely that you'll ever get this error, so it is
-	 * only listed for completeness. */
-	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
-		     _("could not get information about CRTC %d"),
-		     (int) crtc->id);
-	return FALSE;
+        /* Translators: CRTC is a CRT Controller (this is X terminology).
+        * It is *very* unlikely that you'll ever get this error, so it is
+        * only listed for completeness. */
+        g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_RANDR_ERROR,
+                     _("could not get information about CRTC %d"),
+                    (int) crtc->id);
+        return FALSE;
     }
     
     /* GnomeRRMode */
@@ -2364,35 +2532,41 @@ crtc_initialize (GnomeRRCrtc        *crtc,
     a = g_ptr_array_new ();
     for (i = 0; i < info->noutput; ++i)
     {
-	GnomeRROutput *output = gnome_rr_output_by_id (crtc->info, info->outputs[i]);
-	
-	if (output)
-	    g_ptr_array_add (a, output);
+        GnomeRROutput *output = gnome_rr_output_by_id (crtc->info, info->outputs[i]);
+
+        if (output)
+            g_ptr_array_add (a, output);
     }
+
     g_ptr_array_add (a, NULL);
     crtc->current_outputs = (GnomeRROutput **)g_ptr_array_free (a, FALSE);
-    
+
     /* Possible outputs */
     a = g_ptr_array_new ();
     for (i = 0; i < info->npossible; ++i)
     {
-	GnomeRROutput *output = gnome_rr_output_by_id (crtc->info, info->possible[i]);
-	
-	if (output)
-	    g_ptr_array_add (a, output);
+        GnomeRROutput *output = gnome_rr_output_by_id (crtc->info, info->possible[i]);
+
+        if (output)
+            g_ptr_array_add (a, output);
     }
+
     g_ptr_array_add (a, NULL);
     crtc->possible_outputs = (GnomeRROutput **)g_ptr_array_free (a, FALSE);
     
     /* Rotations */
     crtc->current_rotation = gnome_rr_rotation_from_xrotation (info->rotation);
     crtc->rotations = gnome_rr_rotation_from_xrotation (info->rotations);
-    
+    crtc->scale = get_crtc_scale (crtc);
+
     XRRFreeCrtcInfo (info);
 
     /* get an store gamma size */
     crtc->gamma_size = XRRGetCrtcGammaSize (DISPLAY (crtc), crtc->id);
 
+    g_debug ("Initialized GnomeCrtc: %d, at %d, %d, with a scale factor of %.2f (%u global scale)",
+             (int) crtc->id, crtc->x, crtc->y, crtc->scale, gnome_rr_screen_get_global_scale (NULL));
+
     return TRUE;
 }
 
@@ -2430,6 +2604,13 @@ gnome_rr_mode_get_width (GnomeRRMode *mode)
     return mode->width;
 }
 
+guint
+gnome_rr_mode_get_height (GnomeRRMode *mode)
+{
+    g_return_val_if_fail (mode != NULL, 0);
+    return mode->height;
+}
+
 int
 gnome_rr_mode_get_freq (GnomeRRMode *mode)
 {
@@ -2437,11 +2618,27 @@ gnome_rr_mode_get_freq (GnomeRRMode *mode)
     return (mode->freq) / 1000;
 }
 
-guint
-gnome_rr_mode_get_height (GnomeRRMode *mode)
+double
+gnome_rr_mode_get_freq_f (GnomeRRMode *mode)
 {
-    g_return_val_if_fail (mode != NULL, 0);
-    return mode->height;
+    g_return_val_if_fail (mode != NULL, 0.0);
+    return (mode->freq) / 1000.0;
+}
+
+void
+gnome_rr_mode_get_flags            (GnomeRRMode           *mode,
+                                    gboolean              *doublescan,
+                                    gboolean              *interlaced,
+                                    gboolean              *vsync)
+{
+    g_return_if_fail (mode != NULL);
+
+    if (doublescan)
+        *doublescan = mode->doublescan;
+    if (interlaced)
+        *interlaced = mode->interlaced;
+    if (vsync)
+        *vsync = mode->vsync;
 }
 
 static void
@@ -2453,7 +2650,10 @@ mode_initialize (GnomeRRMode *mode, XRRModeInfo *info)
     mode->name = g_strdup (info->name);
     mode->width = info->width;
     mode->height = info->height;
-    mode->freq = ((info->dotClock / (double)info->hTotal) / info->vTotal + 0.5) * 1000;
+    mode->freq = info->dotClock / ((float)info->hTotal * info->vTotal) * 1000;
+    mode->doublescan = (info->modeFlags & RR_DoubleScan);
+    mode->interlaced = (info->modeFlags & RR_Interlace);
+    mode->vsync = (info->modeFlags & RR_VSyncPositive);
 }
 
 static GnomeRRMode *
@@ -2549,3 +2749,271 @@ gnome_rr_crtc_get_gamma (GnomeRRCrtc *crtc, int *size,
     return TRUE;
 }
 
+#define SCALE_FACTORS_PER_INTEGER 4
+#define SCALE_FACTORS_STEPS (1.0 / (float) SCALE_FACTORS_PER_INTEGER)
+
+/* The minimum resolution at which we turn on a window-scale of 2 */
+#define HIDPI_LIMIT 192
+#define HIDPI_MIN_SCALED_HEIGHT 720
+
+/* The minimum screen height at which we turn on a window-scale of 2;
+ * below this there just isn't enough vertical real estate for GNOME
+ * apps to work, and it's better to just be tiny */
+#define HIDPI_MIN_HEIGHT 1500
+
+static gboolean
+is_logical_size_large_enough (int width,
+                              int height)
+{
+  return height > HIDPI_MIN_SCALED_HEIGHT;
+}
+
+static gboolean
+is_scale_valid_for_size (float width,
+                         float height,
+                         float scale)
+{
+  return scale >= MINIMUM_LOGICAL_SCALE_FACTOR &&
+         scale <= MAXIMUM_LOGICAL_SCALE_FACTOR;
+}
+
+static float
+get_closest_scale_factor_for_resolution (float width,
+                                         float height,
+                                         float scale)
+{
+  unsigned int i, j;
+  float scaled_h;
+  float scaled_w;
+  float best_scale;
+  int base_scaled_w;
+  gboolean found_one;
+
+  best_scale = 0;
+
+  if (!is_scale_valid_for_size (width, height, scale))
+    goto out;
+
+  if (fmodf (width, scale) == 0.0 && fmodf (height, scale) == 0.0)
+    return scale;
+
+  i = 0;
+  found_one = FALSE;
+  base_scaled_w = floorf (width / scale);
+  do
+    {
+      for (j = 0; j < 2; j++)
+        {
+          float current_scale;
+          int offset = i * (j ? 1 : -1);
+
+          scaled_w = base_scaled_w + offset;
+          current_scale = width / scaled_w;
+          scaled_h = height / current_scale;
+
+          if (current_scale >= scale + SCALE_FACTORS_STEPS ||
+              current_scale <= scale - SCALE_FACTORS_STEPS ||
+              current_scale < MINIMUM_LOGICAL_SCALE_FACTOR ||
+              current_scale > MAXIMUM_LOGICAL_SCALE_FACTOR)
+            {
+              goto out;
+            }
+
+          if (floorf (scaled_h) == scaled_h)
+            {
+              found_one = TRUE;
+
+              if (fabsf (current_scale - scale) < fabsf (best_scale - scale))
+                best_scale = current_scale;
+            }
+        }
+
+      i++;
+    }
+  while (!found_one);
+
+out:
+  return best_scale;
+}
+
+float *
+gnome_rr_screen_calculate_supported_scales (GnomeRRScreen     *screen,
+                                            int                width,
+                                            int                height,
+                                            int               *n_supported_scales)
+{
+  unsigned int i, j;
+  GArray *supported_scales;
+
+  supported_scales = g_array_new (FALSE, FALSE, sizeof (float));
+
+  for (i = floorf (MINIMUM_LOGICAL_SCALE_FACTOR);
+       i <= ceilf (MAXIMUM_LOGICAL_SCALE_FACTOR);
+       i++)
+    {
+      for (j = 0; j < SCALE_FACTORS_PER_INTEGER; j++)
+        {
+          float scale;
+          float scale_value = i + j * SCALE_FACTORS_STEPS;
+
+          scale = get_closest_scale_factor_for_resolution (width,
+                                                           height,
+                                                           scale_value);
+          if (scale > 0.0f)
+            g_array_append_val (supported_scales, scale);
+        }
+    }
+
+  if (supported_scales->len == 0)
+    {
+      float fallback_scale;
+
+      fallback_scale = MINIMUM_LOGICAL_SCALE_FACTOR;
+      g_array_append_val (supported_scales, fallback_scale);
+    }
+
+  *n_supported_scales = supported_scales->len;
+  return (float *) g_array_free (supported_scales, FALSE);
+}
+
+static void
+get_real_monitor_size (GnomeRRScreen *screen,
+                       GdkMonitor    *monitor,
+                       gint           monitor_index,
+                       gint          *width,
+                       gint          *height)
+{
+    GnomeRROutput **outputs;
+    XID output_id;
+    gint i;
+
+    *width = 0;
+    *height = 0;
+
+    output_id = gdk_x11_screen_get_monitor_output (gdk_screen_get_default (), monitor_index);
+    outputs = gnome_rr_screen_list_outputs (screen);
+
+    for (i = 0; outputs[i] != NULL; i++)
+    {
+        GnomeRROutput *output = outputs[i];
+
+        if (gnome_rr_output_get_id (output) == output_id)
+        {
+            GnomeRRMode *mode = gnome_rr_output_get_current_mode (output);
+
+            if (mode == NULL)
+            {
+                mode = gnome_rr_output_get_preferred_mode (output);
+            }
+
+            if (mode != NULL)
+            {
+                *width = gnome_rr_mode_get_width (mode);
+                *height = gnome_rr_mode_get_height (mode);
+            }
+
+            break;
+        }
+    }
+
+    if (*width == 0 || *height == 0)
+    {
+        GdkRectangle rect;
+        gdk_monitor_get_geometry (monitor, &rect);
+
+        *width = rect.width;
+        *height = rect.height;
+    }
+}
+
+guint
+gnome_rr_screen_calculate_best_global_scale (GnomeRRScreen *screen,
+                                             gint           index)
+{
+    guint window_scale;
+    GdkDisplay *display;
+    GdkMonitor *monitor;
+    int width_mm, height_mm;
+    int monitor_scale;
+    double dpi_x, dpi_y;
+    int real_width, real_height;
+
+    display = gdk_display_get_default ();
+    
+    if (index == -1)
+    {
+        monitor = gdk_display_get_primary_monitor (display);
+
+        index = 0;
+    }
+    else
+    {
+        if ((index >= 0) && (index < gdk_display_get_n_monitors (display)))
+        {
+            monitor = gdk_display_get_monitor (display, index);
+        }
+        else
+        {
+            g_warning ("Invalid monitor index provided (%d)", index);
+            return 1;
+        }
+    }
+
+    /* GdkMonitor geometry is affected by any active current transformation/scaling!!!!
+     * So if I have 2x global scale, and a transform on a 1080 monitor down to 540, that's
+     * what gdk_monitor_get_geometry() returns.  We actually have to get the dimensions from
+     * the current RRMode for the given monitor.
+     */
+    get_real_monitor_size (screen, monitor, index, &real_width, &real_height);
+
+
+    width_mm = gdk_monitor_get_width_mm (monitor);
+    height_mm = gdk_monitor_get_height_mm (monitor);
+    monitor_scale = gdk_monitor_get_scale_factor (monitor);
+
+    window_scale = 1;
+
+    g_debug ("Calculating best global scale for monitor %d. Physical size: %dmm x %dmm,"
+             " REAL pixel size: %d x %d.  Current global scale: %d, reported monitor scale: %d",
+             index, width_mm, height_mm, real_width, real_height, gnome_rr_screen_get_global_scale (NULL), monitor_scale);
+
+    if (real_height < HIDPI_MIN_HEIGHT)
+    {
+        g_debug ("REAL height of %d for monitor %d is less than %d, so the recommended scale will be 1", real_height, index, HIDPI_MIN_HEIGHT);
+        goto out;
+    }
+
+    /* Some monitors/TV encode the aspect ratio (16/9 or 16/10) instead of the physical size */
+    if ((width_mm == 160 && height_mm == 90) ||
+        (width_mm == 160 && height_mm == 100) ||
+        (width_mm == 16 && height_mm == 9) ||
+        (width_mm == 16 && height_mm == 10) ||
+        (width_mm == 0 || height_mm == 0))
+    {
+        g_debug ("Aspect ratio instead of physical dimensions were encoded as the physical size, or the physical"
+                 " size was not set. Unable to reliably calculate the recommended scale, returning 1");
+        goto out;
+    }
+
+    if (width_mm > 0 && height_mm > 0) {
+            dpi_x = (double) real_width / (width_mm / 25.4);
+            dpi_y = (double) real_height / (height_mm / 25.4);
+            /* We don't completely trust these values so both must be high, and never pick higher ratio than
+              2 automatically */
+            if (dpi_x > HIDPI_LIMIT && dpi_y > HIDPI_LIMIT)
+            {
+                g_debug ("The REAL monitor DPI of %.1f x %.1f exceeds the cutoff of %d x %d, recommended scale will be 2",
+                         dpi_x, dpi_y, HIDPI_LIMIT, HIDPI_LIMIT);
+
+                window_scale = 2;
+            }
+            else
+            {
+                g_debug ("The REAL monitor DPI of %.1f x %.1f does not meet the cutoff of %d x %d, recommended scale will be 1",
+                         dpi_x, dpi_y, HIDPI_LIMIT, HIDPI_LIMIT);
+            }
+    }
+
+out:
+    return window_scale;
+}
